# This file is part of VBAO.
#
# VBAO is free software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# VBAO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with VBAO.
# If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import *

from .base import *
from .mixin import PropertyMixin, CommandMixin

DictCons = dict


def use_easydict(b=True):
    global DictCons
    if b:
        from easydict import EasyDict
        DictCons = EasyDict
    else:
        DictCons = dict


class Model(PropertyMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prop_notice = PropertyNotifier()

        # self.property will conflict with Qt
        # None here means it should be set by App.bind()
        self.properties: Dict[str, Any] = None

    def addPropertyListener_from_vm(self, listener: PropertyListenerBase):
        # NotificationHolder can check validity, but we want to provide better error info
        if not isinstance(listener, PropertyListenerBase):
            raise TypeError(f"expect PropertyListenerBase, but get {type(listener)}")
        self._prop_notice.addNotification(listener)

    def triggerPropertyNotifications(self, name: str):
        self._prop_notice.triggerPropertyNotifications(name)


class ViewModel(PropertyMixin, CommandMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prop_notice = PropertyNotifier()
        self._cmd_notice = CommandNotifier()
        self._prop_listener = None
        # Viewmodel contains Model, the result of running “commands” of Model can be directly seen
        self.model = None  # lots of operations will be done with Model, model is better than _model

        self.commands: Dict[str, CommandBase] = DictCons()
        self.properties: Dict[str, Any] = DictCons()

    def addPropertyListener_from_view(self, listener: PropertyListenerBase):
        if not isinstance(listener, PropertyListenerBase):
            raise TypeError(f"expect PropertyListenerBase, but get {type(listener)}")
        self._prop_notice.addNotification(listener)

    def triggerPropertyNotifications(self, name: str):
        self._prop_notice.triggerPropertyNotifications(name)

    def addCommandListener_from_view(self, listener: CommandListenerBase):
        if not isinstance(listener, CommandListenerBase):
            raise TypeError(f"expect CommandListenerBase, but get {type(listener)}")
        self._cmd_notice.addNotification(listener)

    def triggerCommandNotifications(self, name: str, success: bool):
        self._cmd_notice.triggerCommandNotifications(name, success)

    # extension
    def bindModel(self, model: Model, *, verbose=True):
        self.model = model
        model.properties = self.properties
        if self._prop_listener is None:
            if verbose:
                logging.warning("You are binding Model to VM without a listener, set verbose=True to suppress warning.")
        else:
            model.addPropertyListener_from_vm(self._prop_listener)

    @property
    def isModelSet(self):
        return self.model is not None

    def setListener(self, listener: PropertyListenerBase):
        self._prop_listener = listener

    def runCommand(self, cmd_name: str):
        """
        Deprecated. Prefer to run command in View.
        """
        if cmd_name not in self.commands.keys():
            raise ValueError(f"Trying to call an invalid command {cmd_name}. "
                             f"Candidates are: {self.commands.keys()}")

        self.commands.get(cmd_name).execute()


class View(PropertyMixin, CommandMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prop_listener = None
        self.cmd_listener = None

        # None here means it should be set by App.bind()
        self.commands: Dict[str, CommandBase] = None
        self.properties: Dict[str, Any] = None

    def runCommand(self, cmd_name: str):
        if cmd_name not in self.commands.keys():
            raise ValueError(f"Trying to call an invalid command {cmd_name}. "
                             f"Candidates are: {self.commands.keys()}")

        self.commands.get(cmd_name).execute()


class App:
    @classmethod
    def bind(cls, model: Model, viewmodel: ViewModel, view: View,
             bind_vm_n_model=False, *, debug_set_vm_in_view=False):
        if bind_vm_n_model:
            viewmodel.bindModel(model)
        if debug_set_vm_in_view:
            view.viewmodel = viewmodel

        # prop bindings
        # model.properties = viewmodel.properties # done in bindModel
        view.properties = viewmodel.properties

        # cmd bindings
        view.commands = viewmodel.commands

        # event notification
        # model.addPropertyListener_from_vm(viewmodel._prop_listener) # done in bindModel
        viewmodel.addPropertyListener_from_view(view.prop_listener)
        viewmodel.addCommandListener_from_view(view.cmd_listener)



# This file is part of VBAO.
#
# VBAO is free software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# VBAO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with VBAO.
# If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod


class CommandBase(ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # predefine None. For convenience of judging whether set parameter
        self.args = None
        self.kwargs = None

    def setParameter(self, *args, **kwargs):
        """
        Set parameters in self.args & self.kwargs
        """
        self.args = args
        self.kwargs = kwargs

    @abstractmethod
    def execute(self):
        """
        Prefer call onCommandComplete(self, cmd_name: str, success: bool), other than return bool
        """
        pass


class PropertyListenerBase(ABC):
    def __init__(self, to_whom):
        self.master = to_whom

    @abstractmethod
    def onPropertyChanged(self, prop_name: str):
        pass


class CommandListenerBase(ABC):
    def __init__(self, to_whom):
        self.master = to_whom

    @abstractmethod
    def onCommandComplete(self, cmd_name: str, success: bool):
        pass


class NotificationHolder:
    def __init__(self):
        self.arr = []

    def addNotification(self, input):
        if input is None:
            return
        elif isinstance(input, (PropertyListenerBase, CommandListenerBase)):
            self.arr.append(input)
        else:
            print("Not invalid type",input)

    def removeNotification(self, x):
        if isinstance(input, (PropertyListenerBase, CommandListenerBase)):
            self.arr.remove(x)

    def clear(self):
        self.arr.clear()


class PropertyNotifier(NotificationHolder):
    def triggerPropertyNotifications(self, name):
        for listener in self.arr:
            listener.onPropertyChanged(name)


class CommandNotifier(NotificationHolder):
    def triggerCommandNotifications(self, name, success):
        for listener in self.arr:
            listener.onCommandComplete(name, success)


# This file is part of VBAO.
#
# VBAO is free software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# VBAO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with VBAO.
# If not, see <https://www.gnu.org/licenses/>.

from abc import ABC
from functools import wraps
from typing import final

from .base import CommandBase, CommandListenerBase, PropertyListenerBase
from .core import ViewModel


@final
class DummyPropListener(PropertyListenerBase):
    # No need to take args, except for `to_whom` to keep compatible with normal listeners
    # Throws as soon as possible
    @wraps(PropertyListenerBase.__init__)
    def __init__(self, parent=None):
        super().__init__(parent)

    def onPropertyChanged(self, prop_name: str):
        raise NotImplementedError("You should not call on a DummyPropListener")


@final
class DummyCmdListener(CommandListenerBase):
    # No need to take args, except for `to_whom` to keep compatible with normal listeners
    # Throws as soon as possible
    @wraps(CommandListenerBase.__init__)
    def __init__(self, parent=None):
        super().__init__(parent)

    def onCommandComplete(self, cmd_name: str, success: bool):
        raise NotImplementedError("You should not call on a DummyCmdListener")


class CommandDirectCallMixin(CommandBase, ABC):
    """
    can call self.directCall() with parameter in 1 step
    Unfortunately, without override, in `self.execute` you still need to get the args by `self.args` and `self.kwargs`
    """

    def directCall(self, *args, **kwargs):
        self.setParameter(*args, **kwargs)
        self.execute()


class CommandBaseWithOwner(CommandBase, ABC):
    """
    __init__ receives an argument and stores it in self.owner.
    """

    def __init__(self, owner: ViewModel, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.owner: ViewModel = owner


class CommandBaseWithDict(CommandBase, ABC):
    """
    __init__ receives a dict and stores it in self.dict. Allows None.
    """

    def __init__(self, dict_, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if dict_ is not None and not isinstance(dict_, dict):
            raise TypeError(f"CommandBaseWithDict() expects a dict argument but receives {type(dict_)}")
        self.dict = dict_


class LambdaCommand(CommandBase):
    """
    stores a function, and `self.execute` will call it with args passed by `self.setParameter`
    """

    def __init__(self, function, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not callable(function):
            raise TypeError(f"LambdaCommand() expects a function argument but receives {type(function)}")
        self.fn = function

    def execute(self):
        return self.fn(*self.args, **self.kwargs)



from .base import CommandBase, CommandListenerBase, PropertyListenerBase
from .core import Model, ViewModel, View, App
from .core import use_easydict
from .config import qt_installed
from .convenient_class import *




# This file is part of VBAO.
#
# VBAO is free software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# VBAO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with VBAO.
# If not, see <https://www.gnu.org/licenses/>.

from typing import *
from functools import wraps

from .config import qt_installed
from .base import CommandBase


# avoid_qt = True
# keep_old_alias = False

def addDoc(function):
    s = ('You may see a function with suffix _vbao, which is the same as this one, '
         'and is designed to avoid name collision with Qt, etc.')
    if function.__doc__ is None:
        function.__doc__ = s
    else:
        function.__doc__ += '\n' + s
    return function


# It can be done via an if-statement in class definition
# But by doing so, type hint in PyCharm will be a mess
# So use subclasses to get a better type hint
class DictMixinHelper:
    def __init__(self, obj, attribute: str):
        assert isinstance(attribute, str)
        self.attribute = attribute
        self.obj = obj

    @property
    def _inner_dict(self) -> dict:
        return getattr(self.obj, self.attribute)

    def emplace(self, key: str, value: Any) -> None:
        self._inner_dict[key] = value

    def find(self, key: str, verbose: bool = True) -> bool:
        if key in self._inner_dict:
            return True
        elif verbose:
            print(f'{key} is not a valid {self.attribute}, candidates are {self._inner_dict.keys()}')
            return False

    def getValue(self, key: str) -> Optional[Any]:
        if key in self._inner_dict:
            return self._inner_dict[key]
        else:
            return None


class PropertyMixinImpl:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prop_mixin_helper = DictMixinHelper(self, 'properties')

    @addDoc
    def setProperty(self, key: str, value: Any) -> None:
        return self._prop_mixin_helper.emplace(key, value)

    @addDoc
    def getProperty(self, key: str) -> Optional[Any]:
        return self._prop_mixin_helper.getValue(key)

    @addDoc
    def hasProperty(self, key: str, verbose: bool = True) -> bool:
        return self._prop_mixin_helper.find(key, verbose)


class CommandMixinImpl:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cmd_mixin_helper = DictMixinHelper(self, 'commands')

    # change type annotation
    @addDoc
    def setCommand(self, key: str, value: CommandBase) -> None:
        self._cmd_mixin_helper.emplace(key, value)

    @addDoc
    def getCommand(self, key: str) -> Optional[CommandBase]:
        return self._cmd_mixin_helper.getValue(key)

    @addDoc
    def hasCommand(self, key: str, verbose: bool = True) -> bool:
        return self._cmd_mixin_helper.find(key, verbose)


class PropertyMixinAvoidCollide(PropertyMixinImpl):
    setProperty_vbao = PropertyMixinImpl.setProperty
    getProperty_vbao = PropertyMixinImpl.getProperty
    hasProperty_vbao = PropertyMixinImpl.hasProperty


class CommandMixinAvoidCollide(CommandMixinImpl):
    setCommand_vbao = CommandMixinImpl.setCommand
    getCommand_vbao = CommandMixinImpl.getCommand
    hasCommand_vbao = CommandMixinImpl.hasCommand


# if qt_installed:
#     CommandMixin = CommandMixinAvoidCollide
#     PropertyMixin = PropertyMixinAvoidCollide
# else:
#     CommandMixin = CommandMixinImpl
#     PropertyMixin = PropertyMixinImpl

CommandMixin = CommandMixinAvoidCollide
PropertyMixin = PropertyMixinAvoidCollide



# This file is part of VBAO.
#
# VBAO is free software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# VBAO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with VBAO.
# If not, see <https://www.gnu.org/licenses/>.


qt_installed = None
# find out if Qt is installed because Qt's property will conflict with VBAO
try:
    from PIL import ImageQt
    qt_installed = ImageQt.qt_is_installed
except ImportError:
    pass

if qt_installed is None:
    qt_installed = False
    # PIL uninstalled, manually try import Qt
    for qt_module in ["PySide6", "PyQt6", "PySide2", "PySide2"]:
        try:
            if qt_module == "PySide6":
                import PySide6
            elif qt_module == "PyQt6":
                import PyQt6
            elif qt_module == "PySide2":
                import PySide2
            elif qt_module == "PyQt5":
                import PyQt5
        except ImportError:
            continue

        qt_installed = True
        break
